# -*- python -*-
#
# gtk-doc - GTK DocBook documentation generator.
# Copyright (C) 1998  Damon Chaplin
#               2007-2016  Stefan Sauer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#

"""
The scangobj tool gets information about object hierarchies and signals by
compiling and running a small C program. CFLAGS and LDFLAGS must be set
appropriately before running this script.
"""

import logging
import os
import string
import subprocess
import shlex

from . import common, config


COMMON_INCLUDES = """
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <glib-object.h>
"""

QUERY_CHILD_PROPS_PROTOTYPE = "extern GParamSpec** %s (gpointer class, guint *n_properties);"

QUERY_CHILD_PROPS_CODE = """
    if (!child_prop) {
      properties = %s (class, &n_properties);
      if (properties) {
        child_prop = TRUE;
        continue;
      }
   }
"""

MAIN_CODE = """

#ifdef GTK_IS_WIDGET_CLASS
#include <gtk/gtk.h>
#endif
static GType object_types[${ntypes}];

static GType *
get_object_types (void)
{
  gpointer g_object_class;
  gint i = 0;

${get_types}
  object_types[i] = G_TYPE_INVALID;

  /* reference the GObjectClass to initialize the param spec pool
   * potentially needed by interfaces. See http://bugs.gnome.org/571820 */
  g_object_class = g_type_class_ref (G_TYPE_OBJECT);

  /* Need to make sure all the types are loaded in and initialize
   * their signals and properties.
   */
  for (i=0; object_types[i]; i++) {
    if (G_TYPE_IS_CLASSED (object_types[i]))
      g_type_class_ref (object_types[i]);
    if (G_TYPE_IS_INTERFACE (object_types[i]))
      g_type_default_interface_ref (object_types[i]);
  }

  g_type_class_unref (g_object_class);

  return object_types;
}

/*
 * This uses GObject type functions to output signal prototypes and the object
 * hierarchy.
 */

/* The output files */
const gchar *signals_filename = "${new_signals_filename}";
const gchar *hierarchy_filename = "${new_hierarchy_filename}";
const gchar *interfaces_filename = "${new_interfaces_filename}";
const gchar *prerequisites_filename = "${new_prerequisites_filename}";
const gchar *args_filename = "${new_args_filename}";

static void output_signals (void);
static void output_object_signals (FILE *fp,
                                   GType object_type);
static void output_object_signal (FILE *fp,
                                  const gchar *object_class_name,
                                  guint signal_id);
static const gchar * get_type_name (GType type,
                                    gboolean * is_pointer);
static void output_object_hierarchy (void);
static void output_hierarchy (FILE *fp,
                              GType type,
                              guint level);

static void output_object_interfaces (void);
static void output_interfaces (FILE *fp,
                               GType type);

static void output_interface_prerequisites (void);
static void output_prerequisites (FILE *fp,
                                  GType type);

static void output_args (void);
static void output_object_args (FILE *fp, GType object_type);

int
main (${main_func_params})
{
  ${type_init_func};

  get_object_types ();

  output_signals ();
  output_object_hierarchy ();
  output_object_interfaces ();
  output_interface_prerequisites ();
  output_args ();

  return 0;
}

static void
output_signals (void)
{
  FILE *fp;
  gint i;

  fp = fopen (signals_filename, "w");
  if (fp == NULL) {
    g_warning ("Couldn't open output file: %s : %s", signals_filename, g_strerror(errno));
    return;
  }

  for (i = 0; object_types[i]; i++)
    output_object_signals (fp, object_types[i]);

  fclose (fp);
}

static gint
compare_signals (const void *a, const void *b)
{
  const guint *signal_a = a;
  const guint *signal_b = b;

  return strcmp (g_signal_name (*signal_a), g_signal_name (*signal_b));
}

/* This outputs all the signals of one object. */
static void
output_object_signals (FILE *fp, GType object_type)
{
  const gchar *object_class_name;
  guint *signals, n_signals;
  guint sig;

  if (G_TYPE_IS_INSTANTIATABLE (object_type) ||
      G_TYPE_IS_INTERFACE (object_type)) {

    object_class_name = g_type_name (object_type);

    signals = g_signal_list_ids (object_type, &n_signals);
    qsort (signals, n_signals, sizeof (guint), compare_signals);

    for (sig = 0; sig < n_signals; sig++) {
       output_object_signal (fp, object_class_name, signals[sig]);
    }
    g_free (signals);
  }
}

/* This outputs one signal. */
static void
output_object_signal (FILE *fp,
                      const gchar *object_name,
                      guint signal_id)
{
  GSignalQuery query_info;
  const gchar *type_name, *ret_type, *object_arg, *arg_name;
  gchar *pos, *object_arg_lower;
  gboolean is_pointer;
  gchar buffer[1024];
  guint i, param;
  gint param_num, widget_num, event_num, callback_num;
  gint *arg_num;
  gchar signal_name[128];
  gchar flags[16];

  /*  g_print ("Object: %s Signal: %u\\n", object_name, signal_id);*/

  param_num = 1;
  widget_num = event_num = callback_num = 0;

  g_signal_query (signal_id, &query_info);

  /* Output the signal object type and the argument name. We assume the
   * type is a pointer - I think that is OK. We remove "Gtk" or "Gnome" and
   * convert to lower case for the argument name. */
  pos = buffer;
  sprintf (pos, "%s ", object_name);
  pos += strlen (pos);

  /* Try to come up with a sensible variable name for the first arg
   * It chops off 2 know prefixes :/ and makes the name lowercase
   * It should replace lowercase -> uppercase with '_'
   * GFileMonitor -> file_monitor
   * GIOExtensionPoint -> extension_point
   * GtkTreeView -> tree_view
   * if 2nd char is upper case too
   *   search for first lower case and go back one char
   * else
   *   search for next upper case
   */
  if (!strncmp (object_name, "Gtk", 3))
    object_arg = object_name + 3;
  else if (!strncmp (object_name, "Gnome", 5))
    object_arg = object_name + 5;
  else
    object_arg = object_name;

  object_arg_lower = g_ascii_strdown (object_arg, -1);
  sprintf (pos, "*%s\\n", object_arg_lower);
  pos += strlen (pos);
  if (!strncmp (object_arg_lower, "widget", 6))
    widget_num = 2;
  g_free(object_arg_lower);

  /* Convert signal name to use underscores rather than dashes '-'. */
  strncpy (signal_name, query_info.signal_name, 127);
  signal_name[127] = '\\0';
  for (i = 0; signal_name[i]; i++) {
    if (signal_name[i] == '-')
      signal_name[i] = '_';
  }

  /* Output the signal parameters. */
  for (param = 0; param < query_info.n_params; param++) {
    type_name = get_type_name (query_info.param_types[param] & ~G_SIGNAL_TYPE_STATIC_SCOPE, &is_pointer);

    /* Most arguments to the callback are called "arg1", "arg2", etc.
       GtkWidgets are called "widget", "widget2", ...
       GtkCallbacks are called "callback", "callback2", ... */
    if (!strcmp (type_name, "GtkWidget")) {
      arg_name = "widget";
      arg_num = &widget_num;
    }
    else if (!strcmp (type_name, "GtkCallback")
             || !strcmp (type_name, "GtkCCallback")) {
      arg_name = "callback";
      arg_num = &callback_num;
    }
    else {
      arg_name = "arg";
      arg_num = &param_num;
    }
    sprintf (pos, "%s ", type_name);
    pos += strlen (pos);

    if (!arg_num || *arg_num == 0)
      sprintf (pos, "%s%s\\n", is_pointer ? "*" : " ", arg_name);
    else
      sprintf (pos, "%s%s%i\\n", is_pointer ? "*" : " ", arg_name,
               *arg_num);
    pos += strlen (pos);

    if (arg_num) {
      if (*arg_num == 0)
        *arg_num = 2;
      else
        *arg_num += 1;
    }
  }

  pos = flags;
  /* We use one-character flags for simplicity. */
  if (query_info.signal_flags & G_SIGNAL_RUN_FIRST)
    *pos++ = 'f';
  if (query_info.signal_flags & G_SIGNAL_RUN_LAST)
    *pos++ = 'l';
  if (query_info.signal_flags & G_SIGNAL_RUN_CLEANUP)
    *pos++ = 'c';
  if (query_info.signal_flags & G_SIGNAL_NO_RECURSE)
    *pos++ = 'r';
  if (query_info.signal_flags & G_SIGNAL_DETAILED)
    *pos++ = 'd';
  if (query_info.signal_flags & G_SIGNAL_ACTION)
    *pos++ = 'a';
  if (query_info.signal_flags & G_SIGNAL_NO_HOOKS)
    *pos++ = 'h';
  *pos = 0;

  /* Output the return type and function name. */
  ret_type = get_type_name (query_info.return_type & ~G_SIGNAL_TYPE_STATIC_SCOPE, &is_pointer);

  fprintf (fp,
           "<SIGNAL>\\n<NAME>%s::%s</NAME>\\n<RETURNS>%s%s</RETURNS>\\n<FLAGS>%s</FLAGS>\\n%s</SIGNAL>\\n\\n",
           object_name, query_info.signal_name, ret_type, is_pointer ? "*" : "", flags, buffer);
}


/* Returns the type name to use for a signal argument or return value, given
   the GtkType from the signal info. It also sets is_pointer to TRUE if the
   argument needs a '*' since it is a pointer. */
static const gchar *
get_type_name (GType type, gboolean * is_pointer)
{
  const gchar *type_name;

  *is_pointer = FALSE;
  type_name = g_type_name (type);

  switch (type) {
  case G_TYPE_NONE:
  case G_TYPE_CHAR:
  case G_TYPE_UCHAR:
  case G_TYPE_BOOLEAN:
  case G_TYPE_INT:
  case G_TYPE_UINT:
  case G_TYPE_LONG:
  case G_TYPE_ULONG:
  case G_TYPE_FLOAT:
  case G_TYPE_DOUBLE:
  case G_TYPE_POINTER:
    /* These all have normal C type names so they are OK. */
    return type_name;

  case G_TYPE_STRING:
    /* A GtkString is really a gchar*. */
    *is_pointer = TRUE;
    return "gchar";

  case G_TYPE_ENUM:
  case G_TYPE_FLAGS:
    /* We use a gint for both of these. Hopefully a subtype with a decent
       name will be registered and used instead, as GTK+ does itself. */
    return "gint";

  case G_TYPE_BOXED:
    /* The boxed type shouldn't be used itself, only subtypes. Though we
       return 'gpointer' just in case. */
    return "gpointer";

  case G_TYPE_PARAM:
    /* A GParam is really a GParamSpec*. */
    *is_pointer = TRUE;
    return "GParamSpec";

#if GLIB_CHECK_VERSION (2, 25, 9)
  case G_TYPE_VARIANT:
    *is_pointer = TRUE;
    return "GVariant";
#endif

default:
    break;
  }

  /* For all GObject subclasses we can use the class name with a "*",
     e.g. 'GtkWidget *'. */
  if (g_type_is_a (type, G_TYPE_OBJECT))
    *is_pointer = TRUE;

  /* Also catch non GObject root types */
  if (G_TYPE_IS_CLASSED (type))
    *is_pointer = TRUE;

  /* All boxed subtypes will be pointers as well. */
  /* Exception: GStrv */
  if (g_type_is_a (type, G_TYPE_BOXED) &&
      !g_type_is_a (type, G_TYPE_STRV))
    *is_pointer = TRUE;

  /* All pointer subtypes will be pointers as well. */
  if (g_type_is_a (type, G_TYPE_POINTER))
    *is_pointer = TRUE;

  /* But enums are not */
  if (g_type_is_a (type, G_TYPE_ENUM) ||
      g_type_is_a (type, G_TYPE_FLAGS))
    *is_pointer = FALSE;

  return type_name;
}


/* This outputs the hierarchy of all objects which have been initialized,
   i.e. by calling their XXX_get_type() initialization function. */
static void
output_object_hierarchy (void)
{
  FILE *fp;
  gint i,j;
  GType root, type;
  GType root_types[${ntypes}] = { G_TYPE_INVALID, };

  fp = fopen (hierarchy_filename, "w");
  if (fp == NULL) {
    g_warning ("Couldn't open output file: %s : %s", hierarchy_filename, g_strerror(errno));
    return;
  }
  output_hierarchy (fp, G_TYPE_OBJECT, 0);
  output_hierarchy (fp, G_TYPE_INTERFACE, 0);

  for (i=0; object_types[i]; i++) {
    root = object_types[i];
    while ((type = g_type_parent (root))) {
      root = type;
    }
    if ((root != G_TYPE_OBJECT) && (root != G_TYPE_INTERFACE)) {
      for (j=0; root_types[j]; j++) {
        if (root == root_types[j]) {
          root = G_TYPE_INVALID; break;
        }
      }
      if(root) {
        root_types[j] = root;
        output_hierarchy (fp, root, 0);
      }
    }
  }

  fclose (fp);
}

/* This is called recursively to output the hierarchy of a object. */
static void
output_hierarchy (FILE  *fp,
                  GType  type,
                  guint   level)
{
  guint i;
  GType *children;
  guint n_children;

  if (!type)
    return;

  for (i = 0; i < level; i++)
    fprintf (fp, "  ");
  fprintf (fp, "%s\\n", g_type_name (type));

  children = g_type_children (type, &n_children);

  for (i=0; i < n_children; i++)
    output_hierarchy (fp, children[i], level + 1);

  g_free (children);
}

static void output_object_interfaces (void)
{
  guint i;
  FILE *fp;

  fp = fopen (interfaces_filename, "w");
  if (fp == NULL) {
    g_warning ("Couldn't open output file: %s : %s", interfaces_filename, g_strerror(errno));
    return;
  }
  output_interfaces (fp, G_TYPE_OBJECT);

  for (i = 0; object_types[i]; i++) {
    if (!g_type_parent (object_types[i]) &&
        (object_types[i] != G_TYPE_OBJECT) &&
        G_TYPE_IS_INSTANTIATABLE (object_types[i])) {
      output_interfaces (fp, object_types[i]);
    }
  }
  fclose (fp);
}

static void
output_interfaces (FILE  *fp,
                   GType  type)
{
  guint i;
  GType *children, *interfaces;
  guint n_children, n_interfaces;

  if (!type)
    return;

  interfaces = g_type_interfaces (type, &n_interfaces);

  if (n_interfaces > 0) {
    fprintf (fp, "%s", g_type_name (type));
    for (i=0; i < n_interfaces; i++)
        fprintf (fp, " %s", g_type_name (interfaces[i]));
    fprintf (fp, "\\n");
   }
  g_free (interfaces);

  children = g_type_children (type, &n_children);

  for (i=0; i < n_children; i++)
    output_interfaces (fp, children[i]);

  g_free (children);
}

static void output_interface_prerequisites (void)
{
  FILE *fp;

  fp = fopen (prerequisites_filename, "w");
  if (fp == NULL) {
    g_warning ("Couldn't open output file: %s : %s", prerequisites_filename, g_strerror(errno));
    return;
  }
  output_prerequisites (fp, G_TYPE_INTERFACE);
  fclose (fp);
}

static void
output_prerequisites (FILE  *fp,
                      GType  type)
{
#if GLIB_CHECK_VERSION(2,1,0)
  guint i;
  GType *children, *prerequisites;
  guint n_children, n_prerequisites;

  if (!type)
    return;

  prerequisites = g_type_interface_prerequisites (type, &n_prerequisites);

  if (n_prerequisites > 0) {
    fprintf (fp, "%s", g_type_name (type));
    for (i=0; i < n_prerequisites; i++)
        fprintf (fp, " %s", g_type_name (prerequisites[i]));
    fprintf (fp, "\\n");
   }
  g_free (prerequisites);

  children = g_type_children (type, &n_children);

  for (i=0; i < n_children; i++)
    output_prerequisites (fp, children[i]);

  g_free (children);
#endif
}

static void
output_args (void)
{
  FILE *fp;
  gint i;

  fp = fopen (args_filename, "w");
  if (fp == NULL) {
    g_warning ("Couldn't open output file: %s : %s", args_filename, g_strerror(errno));
    return;
  }

  for (i = 0; object_types[i]; i++) {
    output_object_args (fp, object_types[i]);
  }

  fclose (fp);
}

static gint
compare_param_specs (const void *a, const void *b)
{
  GParamSpec *spec_a = *(GParamSpec **)a;
  GParamSpec *spec_b = *(GParamSpec **)b;

  return strcmp (g_param_spec_get_name (spec_a), g_param_spec_get_name (spec_b));
}

/* Its common to have unsigned properties restricted
 * to the signed range. Therefore we make this look
 * a bit nicer by spelling out the max constants.
 */

/* Don't use "==" with floats, it might trigger a gcc warning.  */
#define GTKDOC_COMPARE_FLOAT(x, y) (x <= y && x >= y)

static gchar*
describe_double_constant (gdouble value)
{
  gchar *desc;

  if (GTKDOC_COMPARE_FLOAT (value, G_MAXDOUBLE))
    desc = g_strdup ("G_MAXDOUBLE");
  else if (GTKDOC_COMPARE_FLOAT (value, G_MINDOUBLE))
    desc = g_strdup ("G_MINDOUBLE");
  else if (GTKDOC_COMPARE_FLOAT (value, -G_MAXDOUBLE))
    desc = g_strdup ("-G_MAXDOUBLE");
  else if (GTKDOC_COMPARE_FLOAT (value, G_MAXFLOAT))
    desc = g_strdup ("G_MAXFLOAT");
  else if (GTKDOC_COMPARE_FLOAT (value, G_MINFLOAT))
    desc = g_strdup ("G_MINFLOAT");
  else if (GTKDOC_COMPARE_FLOAT (value, -G_MAXFLOAT))
    desc = g_strdup ("-G_MAXFLOAT");
  else{
    /* make sure floats are output with a decimal dot irrespective of
    * current locale. Use formatd since we want human-readable numbers
    * and do not need the exact same bit representation when deserialising */
    desc = g_malloc0 (G_ASCII_DTOSTR_BUF_SIZE);
    g_ascii_formatd (desc, G_ASCII_DTOSTR_BUF_SIZE, "%g", value);
  }

  return desc;
}

static gchar*
describe_signed_constant (gsize size, gint64 value)
{
  gchar *desc = NULL;

  switch (size) {
    case 2:
      if (sizeof (int) == 2) {
        if (value == G_MAXINT)
          desc = g_strdup ("G_MAXINT");
        else if (value == G_MININT)
          desc = g_strdup ("G_MININT");
      }
      break;
    case 4:
      if (sizeof (int) == 4) {
        if (value == G_MAXINT)
          desc = g_strdup ("G_MAXINT");
        else if (value == G_MININT)
          desc = g_strdup ("G_MININT");
      }
      if (value == G_MAXLONG)
        desc = g_strdup ("G_MAXLONG");
      else if (value == G_MINLONG)
        desc = g_strdup ("G_MINLONG");
      break;
    case 8:
      if (value == G_MAXINT64)
        desc = g_strdup ("G_MAXINT64");
      else if (value == G_MININT64)
        desc = g_strdup ("G_MININT64");
      break;
    default:
      break;
  }
  if (!desc)
    desc = g_strdup_printf ("%" G_GINT64_FORMAT, value);

  return desc;
}

static gchar*
describe_unsigned_constant (gsize size, guint64 value)
{
  gchar *desc = NULL;

  switch (size) {
    case 2:
      if (sizeof (int) == 2) {
        if (value == (guint64)G_MAXINT)
          desc = g_strdup ("G_MAXINT");
        else if (value == G_MAXUINT)
          desc = g_strdup ("G_MAXUINT");
      }
      break;
    case 4:
      if (sizeof (int) == 4) {
        if (value == (guint64)G_MAXINT)
          desc = g_strdup ("G_MAXINT");
        else if (value == G_MAXUINT)
          desc = g_strdup ("G_MAXUINT");
      }
      if (desc == NULL) {
        if (value == (guint64)G_MAXLONG)
          desc = g_strdup ("G_MAXLONG");
        else if (value == G_MAXULONG)
          desc = g_strdup ("G_MAXULONG");
      }
      break;
    case 8:
      if (value == G_MAXINT64)
        desc = g_strdup ("G_MAXINT64");
      else if (value == G_MAXUINT64)
        desc = g_strdup ("G_MAXUINT64");
      break;
    default:
      break;
  }
  if (!desc)
    desc = g_strdup_printf ("%" G_GUINT64_FORMAT, value);

  return desc;
}

static gchar*
describe_type (GParamSpec *spec)
{
  gchar *desc;
  gchar *lower;
  gchar *upper;

  if (G_IS_PARAM_SPEC_CHAR (spec)) {
    GParamSpecChar *pspec = G_PARAM_SPEC_CHAR (spec);

    lower = describe_signed_constant (sizeof(gchar), pspec->minimum);
    upper = describe_signed_constant (sizeof(gchar), pspec->maximum);
    if (pspec->minimum == G_MININT8 && pspec->maximum == G_MAXINT8)
      desc = g_strdup ("");
    else if (pspec->minimum == G_MININT8)
      desc = g_strdup_printf ("<= %s", upper);
    else if (pspec->maximum == G_MAXINT8)
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
  else if (G_IS_PARAM_SPEC_UCHAR (spec)) {
    GParamSpecUChar *pspec = G_PARAM_SPEC_UCHAR (spec);

    lower = describe_unsigned_constant (sizeof(guchar), pspec->minimum);
    upper = describe_unsigned_constant (sizeof(guchar), pspec->maximum);
    if (pspec->minimum == 0 && pspec->maximum == G_MAXUINT8)
      desc = g_strdup ("");
    else if (pspec->minimum == 0)
      desc = g_strdup_printf ("<= %s", upper);
    else if (pspec->maximum == G_MAXUINT8)
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
  else if (G_IS_PARAM_SPEC_INT (spec)) {
    GParamSpecInt *pspec = G_PARAM_SPEC_INT (spec);

    lower = describe_signed_constant (sizeof(gint), pspec->minimum);
    upper = describe_signed_constant (sizeof(gint), pspec->maximum);
    if (pspec->minimum == G_MININT && pspec->maximum == G_MAXINT)
      desc = g_strdup ("");
    else if (pspec->minimum == G_MININT)
      desc = g_strdup_printf ("<= %s", upper);
    else if (pspec->maximum == G_MAXINT)
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
  else if (G_IS_PARAM_SPEC_UINT (spec)) {
    GParamSpecUInt *pspec = G_PARAM_SPEC_UINT (spec);

    lower = describe_unsigned_constant (sizeof(guint), pspec->minimum);
    upper = describe_unsigned_constant (sizeof(guint), pspec->maximum);
    if (pspec->minimum == 0 && pspec->maximum == G_MAXUINT)
      desc = g_strdup ("");
    else if (pspec->minimum == 0)
      desc = g_strdup_printf ("<= %s", upper);
    else if (pspec->maximum == G_MAXUINT)
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
  else if (G_IS_PARAM_SPEC_LONG (spec)) {
    GParamSpecLong *pspec = G_PARAM_SPEC_LONG (spec);

    lower = describe_signed_constant (sizeof(glong), pspec->minimum);
    upper = describe_signed_constant (sizeof(glong), pspec->maximum);
    if (pspec->minimum == G_MINLONG && pspec->maximum == G_MAXLONG)
      desc = g_strdup ("");
    else if (pspec->minimum == G_MINLONG)
      desc = g_strdup_printf ("<= %s", upper);
    else if (pspec->maximum == G_MAXLONG)
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
  else if (G_IS_PARAM_SPEC_ULONG (spec)) {
    GParamSpecULong *pspec = G_PARAM_SPEC_ULONG (spec);

    lower = describe_unsigned_constant (sizeof(gulong), pspec->minimum);
    upper = describe_unsigned_constant (sizeof(gulong), pspec->maximum);
    if (pspec->minimum == 0 && pspec->maximum == G_MAXULONG)
      desc = g_strdup ("");
    else if (pspec->minimum == 0)
      desc = g_strdup_printf ("<= %s", upper);
    else if (pspec->maximum == G_MAXULONG)
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
  else if (G_IS_PARAM_SPEC_INT64 (spec)) {
    GParamSpecInt64 *pspec = G_PARAM_SPEC_INT64 (spec);

    lower = describe_signed_constant (sizeof(gint64), pspec->minimum);
    upper = describe_signed_constant (sizeof(gint64), pspec->maximum);
    if (pspec->minimum == G_MININT64 && pspec->maximum == G_MAXINT64)
      desc = g_strdup ("");
    else if (pspec->minimum == G_MININT64)
      desc = g_strdup_printf ("<= %s", upper);
    else if (pspec->maximum == G_MAXINT64)
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
  else if (G_IS_PARAM_SPEC_UINT64 (spec)) {
    GParamSpecUInt64 *pspec = G_PARAM_SPEC_UINT64 (spec);

    lower = describe_unsigned_constant (sizeof(guint64), pspec->minimum);
    upper = describe_unsigned_constant (sizeof(guint64), pspec->maximum);
    if (pspec->minimum == 0 && pspec->maximum == G_MAXUINT64)
      desc = g_strdup ("");
    else if (pspec->minimum == 0)
      desc = g_strdup_printf ("<= %s", upper);
    else if (pspec->maximum == G_MAXUINT64)
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
  else if (G_IS_PARAM_SPEC_FLOAT (spec)) {
    GParamSpecFloat *pspec = G_PARAM_SPEC_FLOAT (spec);

    lower = describe_double_constant (pspec->minimum);
    upper = describe_double_constant (pspec->maximum);
    if (GTKDOC_COMPARE_FLOAT (pspec->minimum, -G_MAXFLOAT)) {
      if (GTKDOC_COMPARE_FLOAT (pspec->maximum, G_MAXFLOAT))
        desc = g_strdup ("");
      else
        desc = g_strdup_printf ("<= %s", upper);
    }
    else if (GTKDOC_COMPARE_FLOAT (pspec->maximum, G_MAXFLOAT))
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
  else if (G_IS_PARAM_SPEC_DOUBLE (spec)) {
    GParamSpecDouble *pspec = G_PARAM_SPEC_DOUBLE (spec);

    lower = describe_double_constant (pspec->minimum);
    upper = describe_double_constant (pspec->maximum);
    if (GTKDOC_COMPARE_FLOAT (pspec->minimum, -G_MAXDOUBLE)) {
      if (GTKDOC_COMPARE_FLOAT (pspec->maximum, G_MAXDOUBLE))
        desc = g_strdup ("");
      else
        desc = g_strdup_printf ("<= %s", upper);
    }
    else if (GTKDOC_COMPARE_FLOAT (pspec->maximum, G_MAXDOUBLE))
      desc = g_strdup_printf (">= %s", lower);
    else
      desc = g_strdup_printf ("[%s,%s]", lower, upper);
    g_free (lower);
    g_free (upper);
  }
#if GLIB_CHECK_VERSION (2, 12, 0)
  else if (G_IS_PARAM_SPEC_GTYPE (spec)) {
    GParamSpecGType *pspec = G_PARAM_SPEC_GTYPE (spec);
    gboolean is_pointer;

    desc = g_strdup (get_type_name (pspec->is_a_type, &is_pointer));
  }
#endif
#if GLIB_CHECK_VERSION (2, 25, 9)
  else if (G_IS_PARAM_SPEC_VARIANT (spec)) {
    GParamSpecVariant *pspec = G_PARAM_SPEC_VARIANT (spec);
    gchar *variant_type;

    variant_type = g_variant_type_dup_string (pspec->type);
    desc = g_strdup_printf ("GVariant<%s>", variant_type);
    g_free (variant_type);
  }
#endif
  else {
    desc = g_strdup ("");
  }

  return desc;
}

static gchar*
describe_default (GParamSpec *spec)
{
  gchar *desc;

  if (G_IS_PARAM_SPEC_CHAR (spec)) {
    GParamSpecChar *pspec = G_PARAM_SPEC_CHAR (spec);

    desc = g_strdup_printf ("%d", pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_UCHAR (spec)) {
    GParamSpecUChar *pspec = G_PARAM_SPEC_UCHAR (spec);

    desc = g_strdup_printf ("%u", pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_BOOLEAN (spec)) {
    GParamSpecBoolean *pspec = G_PARAM_SPEC_BOOLEAN (spec);

    desc = g_strdup_printf ("%s", pspec->default_value ? "TRUE" : "FALSE");
  }
  else if (G_IS_PARAM_SPEC_INT (spec)) {
    GParamSpecInt *pspec = G_PARAM_SPEC_INT (spec);

    desc = g_strdup_printf ("%d", pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_UINT (spec)) {
    GParamSpecUInt *pspec = G_PARAM_SPEC_UINT (spec);

    desc = g_strdup_printf ("%u", pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_LONG (spec)) {
    GParamSpecLong *pspec = G_PARAM_SPEC_LONG (spec);

    desc = g_strdup_printf ("%ld", pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_LONG (spec)) {
    GParamSpecULong *pspec = G_PARAM_SPEC_ULONG (spec);

    desc = g_strdup_printf ("%lu", pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_INT64 (spec)) {
    GParamSpecInt64 *pspec = G_PARAM_SPEC_INT64 (spec);

    desc = g_strdup_printf ("%" G_GINT64_FORMAT, pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_UINT64 (spec))
    {
      GParamSpecUInt64 *pspec = G_PARAM_SPEC_UINT64 (spec);

      desc = g_strdup_printf ("%" G_GUINT64_FORMAT, pspec->default_value);
    }
  else if (G_IS_PARAM_SPEC_UNICHAR (spec)) {
    GParamSpecUnichar *pspec = G_PARAM_SPEC_UNICHAR (spec);

    if (g_unichar_isprint (pspec->default_value))
      desc = g_strdup_printf ("'%c'", pspec->default_value);
    else
      desc = g_strdup_printf ("%u", pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_ENUM (spec)) {
    GParamSpecEnum *pspec = G_PARAM_SPEC_ENUM (spec);

    GEnumValue *value = g_enum_get_value (pspec->enum_class, pspec->default_value);
    if (value)
      desc = g_strdup_printf ("%s", value->value_name);
    else
      desc = g_strdup_printf ("%d", pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_FLAGS (spec)) {
    GParamSpecFlags *pspec = G_PARAM_SPEC_FLAGS (spec);
    guint default_value;
    GString *acc;

    default_value = pspec->default_value;
    acc = g_string_new ("");

    while (default_value) {
      GFlagsValue *value = g_flags_get_first_value (pspec->flags_class, default_value);

      if (!value)
        break;

      if (acc->len > 0)
        g_string_append (acc, " | ");
      g_string_append (acc, value->value_name);

      default_value &= ~value->value;
    }

    if (default_value == 0)
      desc = g_string_free (acc, FALSE);
    else {
      desc = g_strdup_printf ("%d", pspec->default_value);
      g_string_free (acc, TRUE);
    }
  }
  else if (G_IS_PARAM_SPEC_FLOAT (spec)) {
    GParamSpecFloat *pspec = G_PARAM_SPEC_FLOAT (spec);

    /* make sure floats are output with a decimal dot irrespective of
     * current locale. Use formatd since we want human-readable numbers
     * and do not need the exact same bit representation when deserialising */
    desc = g_malloc0 (G_ASCII_DTOSTR_BUF_SIZE);
    g_ascii_formatd (desc, G_ASCII_DTOSTR_BUF_SIZE, "%g",
        pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_DOUBLE (spec)) {
    GParamSpecDouble *pspec = G_PARAM_SPEC_DOUBLE (spec);

    /* make sure floats are output with a decimal dot irrespective of
     * current locale. Use formatd since we want human-readable numbers
     * and do not need the exact same bit representation when deserialising */
    desc = g_malloc0 (G_ASCII_DTOSTR_BUF_SIZE);
    g_ascii_formatd (desc, G_ASCII_DTOSTR_BUF_SIZE, "%g",
        pspec->default_value);
  }
  else if (G_IS_PARAM_SPEC_STRING (spec)) {
    GParamSpecString *pspec = G_PARAM_SPEC_STRING (spec);

    if (pspec->default_value) {
      gchar *esc = g_strescape (pspec->default_value, NULL);
      desc = g_strdup_printf ("\\"%s\\"", esc);
      g_free (esc);
    }
    else
      desc = g_strdup_printf ("NULL");
  }
#if GLIB_CHECK_VERSION (2, 25, 9)
  else if (G_IS_PARAM_SPEC_VARIANT (spec)) {
    GParamSpecVariant *pspec = G_PARAM_SPEC_VARIANT (spec);

    if (pspec->default_value)
      desc = g_variant_print (pspec->default_value, TRUE);
    else
      desc = g_strdup ("NULL");
  }
#endif
  else {
    desc = g_strdup ("");
  }

  return desc;
}


static void
output_object_args (FILE *fp, GType object_type)
{
  gpointer class;
  const gchar *object_class_name;
  guint arg;
  gchar flags[16], *pos;
  GParamSpec **properties;
  guint n_properties;
  gboolean child_prop;
  gboolean style_prop;
  gboolean is_pointer;
  const gchar *type_name;
  gchar *type_desc;
  gchar *default_value;

  if (G_TYPE_IS_OBJECT (object_type)) {
    class = g_type_class_peek (object_type);
    if (!class)
      return;

    properties = g_object_class_list_properties (class, &n_properties);
  }
#if GLIB_MAJOR_VERSION > 2 || (GLIB_MAJOR_VERSION == 2 && GLIB_MINOR_VERSION >= 3)
  else if (G_TYPE_IS_INTERFACE (object_type)) {
    class = g_type_default_interface_ref (object_type);

    if (!class)
      return;

    properties = g_object_interface_list_properties (class, &n_properties);
  }
#endif
  else
    return;

  object_class_name = g_type_name (object_type);

  child_prop = FALSE;
  style_prop = FALSE;

  while (TRUE) {
    qsort (properties, n_properties, sizeof (GParamSpec *), compare_param_specs);
    for (arg = 0; arg < n_properties; arg++) {
      GParamSpec *spec = properties[arg];
      const gchar *nick, *blurb, *dot;

      if (spec->owner_type != object_type)
        continue;

      pos = flags;
      /* We use one-character flags for simplicity. */
      if (child_prop && !style_prop)
           *pos++ = 'c';
      if (style_prop)
           *pos++ = 's';
      if (spec->flags & G_PARAM_READABLE)
         *pos++ = 'r';
      if (spec->flags & G_PARAM_WRITABLE)
        *pos++ = 'w';
      if (spec->flags & G_PARAM_CONSTRUCT)
        *pos++ = 'x';
      if (spec->flags & G_PARAM_CONSTRUCT_ONLY)
        *pos++ = 'X';
      *pos = 0;

      nick = g_param_spec_get_nick (spec);
      blurb = g_param_spec_get_blurb (spec);

      dot = "";
      if (blurb) {
        int str_len = strlen (blurb);
        if (str_len > 0  && blurb[str_len - 1] != '.')
          dot = ".";
      }

      type_desc = describe_type (spec);
      default_value = describe_default (spec);
      type_name = get_type_name (spec->value_type, &is_pointer);
      fprintf (fp, "<ARG>\\n"
                   "<NAME>%s::%s</NAME>\\n"
                   "<TYPE>%s%s</TYPE>\\n"
                   "<RANGE>%s</RANGE>\\n"
                   "<FLAGS>%s</FLAGS>\\n"
                   "<NICK>%s</NICK>\\n"
                   "<BLURB>%s%s</BLURB>\\n"
                   "<DEFAULT>%s</DEFAULT>\\n"
                   "</ARG>\\n\\n",
               object_class_name, g_param_spec_get_name (spec), type_name,
               is_pointer ? "*" : "", type_desc, flags, nick ? nick : "(null)",
               blurb ? blurb : "(null)", dot, default_value);
      g_free (type_desc);
      g_free (default_value);
    }

    g_free (properties);

#ifdef GTK_IS_CONTAINER_CLASS
#if !GTK_CHECK_VERSION(3,96,0)
    if (!child_prop && GTK_IS_CONTAINER_CLASS (class)) {
      properties = gtk_container_class_list_child_properties (class, &n_properties);
      child_prop = TRUE;
      continue;
    }
#endif
#endif

#ifdef GTK_IS_CELL_AREA_CLASS
    if (!child_prop && GTK_IS_CELL_AREA_CLASS (class)) {
      properties = gtk_cell_area_class_list_cell_properties (class, &n_properties);
      child_prop = TRUE;
      continue;
    }
#endif

#ifdef GTK_IS_WIDGET_CLASS
#if GTK_CHECK_VERSION(2,1,0) && !GTK_CHECK_VERSION(3,89,2)
    if (!style_prop && GTK_IS_WIDGET_CLASS (class)) {
      properties = gtk_widget_class_list_style_properties (GTK_WIDGET_CLASS (class), &n_properties);
      style_prop = TRUE;
      continue;
    }
#endif
#endif

"""

MAIN_CODE_END = """
    break;
  }
}
"""


def execute_command(options, description, command, env=None):
    if options.verbose:
        call = subprocess.check_call
    else:
        call = subprocess.check_output

    try:
        call(command, env=env)
    except subprocess.CalledProcessError as e:
        logging.warning('%s scanner failed: %d, command: %s', description,
                        e.returncode, ' '.join(command))
        return e.returncode
    except OSError as e:
        logging.warning('%s scanner failed: %s, command: %s', description,
                        str(e), ' '.join(command))
        return 1
    return 0


def run(options):
    logging.info('options: %s', str(options.__dict__))

    c_file = options.module + '-scan.c'
    output = open(c_file, 'w', encoding='utf-8')

    base_filename = os.path.join(options.output_dir, options.module)
    old_signals_filename = base_filename + '.signals'
    new_signals_filename = base_filename + '.signals.new'
    old_hierarchy_filename = base_filename + '.hierarchy'
    new_hierarchy_filename = base_filename + '.hierarchy.new'
    old_interfaces_filename = base_filename + '.interfaces'
    new_interfaces_filename = base_filename + '.interfaces.new'
    old_prerequisites_filename = base_filename + '.prerequisites'
    new_prerequisites_filename = base_filename + '.prerequisites.new'
    old_args_filename = base_filename + '.args'
    new_args_filename = base_filename + '.args.new'

    # generate a C program to scan the types

    includes = ""
    forward_decls = ""
    get_types = ""
    ntypes = 1

    for line in open(options.types, 'r', encoding='utf-8'):
        if line.startswith('#include'):
            includes += line
        elif line.startswith('%') or line.strip() == '':
            continue
        else:
            line = line.strip()
            get_types += '  object_types[i++] = ' + line + ' ();\n'
            forward_decls += 'extern GType ' + line + ' (void);\n'
            ntypes += 1

    output.write(COMMON_INCLUDES)

    if includes:
        output.write(includes)
    else:
        output.write(forward_decls)

    if options.query_child_properties:
        output.write(QUERY_CHILD_PROPS_PROTOTYPE % options.query_child_properties)

    # substitute local vars in the template
    type_init_func = options.type_init_func
    main_func_params = "int argc, char *argv[]"
    if "argc" in type_init_func and "argv" not in type_init_func:
        main_func_params = "int argc, G_GNUC_UNUSED char *argv[]"
    elif "argc" not in type_init_func and "argv" in type_init_func:
        main_func_params = "G_GNUC_UNUSED int argc, char *argv[]"
    elif "argc" not in type_init_func and "argv" not in type_init_func:
        main_func_params = "void"

    # TODO: we should explicitly reference the vars that we substitute
    # template_vars = {
    #     'get_types': get_types,
    #     'ntypes': ntypes,
    #     'main_func_params', main_func_params,
    #     'new_args_filename': new_args_filename,
    #     'new_hierarchy_filename': new_hierarchy_filename,
    #     'new_interfaces_filename': new_interfaces_filename,
    #     'new_prerequisites_filename': new_prerequisites_filename,
    #     'new_signals_filename': new_signals_filename,
    #     'type_init_func': type_init_func
    # }
    # output.write(string.Template(MAIN_CODE).substitute(template_vars))
    output.write(string.Template(MAIN_CODE).substitute(locals()))

    if options.query_child_properties:
        output.write(QUERY_CHILD_PROPS_CODE % options.query_child_properties)

    output.write(MAIN_CODE_END)

    output.close()

    # Compile and run our file
    if 'libtool' in options.cc:
        o_file = options.module + '-scan.lo'
    else:
        o_file = options.module + '-scan.o'

    x_file = options.module + '-scan' + config.exeext

    logging.debug('Intermediate scanner files: %s, %s, %s', c_file, o_file, x_file)

    res = execute_command(options, 'Compiling',
                          shlex.split(options.cc) + shlex.split(options.cflags) +
                          ["-c", "-o", o_file, c_file])
    if res:
        return res

    res = execute_command(options, 'Linking',
                          shlex.split(options.ld) + [o_file] +
                          shlex.split(options.ldflags) + ['-o', x_file])
    if res:
        return res

    run_env = os.environ.copy()
    run_env['LC_MESSAGES'] = 'C'
    run_env.pop('LC_ALL', None)
    res = execute_command(options, 'Running',
                          shlex.split(options.run) + ['./' + x_file],
                          env=run_env)
    if res:
        return res

    logging.debug('Scan complete')
    if 'GTK_DOC_KEEP_INTERMEDIATE' not in os.environ:
        os.unlink(c_file)
        os.unlink(o_file)
        os.unlink(x_file)
        if 'libtool' in options.cc:
            o_file = options.module + '-scan.o'
            if os.path.exists(o_file):
                os.unlink(o_file)
    else:
        logging.debug('Keeping generated sources for analysis: %s, %s, %s',
                      c_file, o_file, x_file)

    common.UpdateFileIfChanged(old_signals_filename, new_signals_filename, False)
    common.UpdateFileIfChanged(old_hierarchy_filename, new_hierarchy_filename, False)
    common.UpdateFileIfChanged(old_interfaces_filename, new_interfaces_filename, False)
    common.UpdateFileIfChanged(old_prerequisites_filename, new_prerequisites_filename, False)
    common.UpdateFileIfChanged(old_args_filename, new_args_filename, False)

    return 0



# -*- python -*-
#
# gtk-doc - GTK DocBook documentation generator.
# Copyright (C) 2018  Stefan Sauer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#

import textwrap
import unittest

from gtkdoc import mkdb


class ScanSourceContentTestCase(unittest.TestCase):
    """Baseclass for the source scanner tests."""

    def setUp(self):
        mkdb.MODULE = 'test'
        mkdb.SymbolDocs = {}


class ScanSourceContent(ScanSourceContentTestCase):

    def test_EmptyInput(self):
        blocks = mkdb.ScanSourceContent([])
        self.assertEqual(0, len(blocks))

    def test_SkipsSingleLineComment(self):
        blocks = mkdb.ScanSourceContent("/** foo */")
        self.assertEqual(0, len(blocks))

    def test_FindsSingleDocComment(self):
        blocks = mkdb.ScanSourceContent("""\
            /**
             * symbol:
             *
             * Description.
             */""".splitlines(keepends=True))
        self.assertEqual(1, len(blocks))


class ParseCommentBlock(ScanSourceContentTestCase):

    def test_EmptyInput(self):
        mkdb.ParseCommentBlock([])
        self.assertEqual({}, mkdb.SourceSymbolDocs)

    def test_FindsDocComment(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:

             Description.
             """).splitlines(keepends=True))
        self.assertEqual({'symbol': 'Description.\n'}, mkdb.SourceSymbolDocs)

    def test_FindsDocCommentWithParam(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:
             @par: value

             Description.
             """).splitlines(keepends=True))
        self.assertEqual({'symbol': 'Description.\n'}, mkdb.SourceSymbolDocs)
        self.assertIn('symbol', mkdb.SourceSymbolParams)
        self.assertEqual({'par': 'value\n'}, mkdb.SourceSymbolParams['symbol'])

    def test_FindsDocCommentWithMultilineParam(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:
             @par: value docs with
               two lines

             Description.
             """).splitlines(keepends=True))
        self.assertEqual({'symbol': 'Description.\n'}, mkdb.SourceSymbolDocs)
        self.assertIn('symbol', mkdb.SourceSymbolParams)
        self.assertEqual({'par': 'value docs with\ntwo lines\n'}, mkdb.SourceSymbolParams['symbol'])

    def test_FindsDocCommentWithReturns(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:

             Description.

             Returns: result
             """).splitlines(keepends=True))
        # TODO: trim multiple newlines in code
        self.assertEqual({'symbol': 'Description.\n\n'}, mkdb.SourceSymbolDocs)
        self.assertIn('symbol', mkdb.SourceSymbolParams)
        # TODO: trim whitespace in code
        self.assertEqual({'Returns': ' result\n'}, mkdb.SourceSymbolParams['symbol'])

    def test_FindsDocCommentWithSince(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:

             Since: 0.1
             """).splitlines(keepends=True))
        self.assertIn('symbol', mkdb.Since)
        self.assertEqual('0.1', mkdb.Since['symbol'])

    def test_FindsDocCommentWithDeprecated(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:

             Deprecated: use function() instead
             """).splitlines(keepends=True))
        self.assertIn('symbol', mkdb.Deprecated)
        # TODO: trim whitespace in code
        self.assertEqual(' use function() instead\n', mkdb.Deprecated['symbol'])

    def test_FindsDocCommentWithStability(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:

             Stability: stable
             """).splitlines(keepends=True))
        self.assertIn('symbol', mkdb.StabilityLevel)
        self.assertEqual('Stable', mkdb.StabilityLevel['symbol'])

    def test_HandlesHTMLEntities(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:

             < & >.
             """).splitlines(keepends=True))
        self.assertEqual({'symbol': '&lt; &amp; &gt;.\n'}, mkdb.SourceSymbolDocs)


class ParseSectionCommentBlock(ScanSourceContentTestCase):

    def test_FindsSectionBlock(self):
        # TODO: maybe override common.LogWarning() instead and capture the messages
        # Suppress: 'Section symbol is not defined in the test-sections.txt file'
        mkdb.KnownSymbols['symbol:long_description'] = 1
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             SECTION:symbol
             @short_description: short module description

             Module description.
             """).splitlines(keepends=True))
        self.assertIn('symbol:short_description', mkdb.SourceSymbolDocs)
        self.assertEqual('short module description\n', mkdb.SourceSymbolDocs['symbol:short_description'])
        self.assertIn('symbol:long_description', mkdb.SourceSymbolDocs)
        self.assertEqual('Module description.\n', mkdb.SourceSymbolDocs['symbol:long_description'])

    # TODO(ensonic): we need to refactor the code first (see comment there)
    # def test_FindsProgramBlock(self):
    #     mkdb.ParseCommentBlock(textwrap.dedent("""\
    #         PROGRAM:symbol
    #         @short_description: short program description
    #         @synopsis: test-program [*OPTIONS*...] --arg1 *arg* *FILE*
    #         @see_also: test(1)
    #         @--arg1 *arg*: set arg1 to *arg*
    #         @-v, --version: Print the version number
    #         @-h, --help: Print the help message
    #
    #         Program description.
    #          """).splitlines(keepends=True))
    #     self.assertIn('symbol:short_description', mkdb.SourceSymbolDocs)
    #     self.assertEqual('short program description\n', mkdb.SourceSymbolDocs['symbol:short_description'])
    #     self.assertIn('symbol:long_description', mkdb.SourceSymbolDocs)
    #     self.assertEqual('Program description.\n', mkdb.SourceSymbolDocs['symbol:long_description'])


class ScanSourceContentAnnotations(ScanSourceContentTestCase):

    def test_ParamAnnotation(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:
             @par: (allow-none): value

             description.
             """).splitlines(keepends=True))
        # TODO: we only extract those when outputting docbook, thats silly
        # self.assertEqual({'par': 'value\n'}, mkdb.SourceSymbolParams['symbol'])
        self.assertEqual({}, mkdb.SymbolAnnotations)

    def test_RetunsAnnotation(self):
        mkdb.ParseCommentBlock(textwrap.dedent("""\
             symbol:

             description.

             Returns: (transfer full) result.
             """).splitlines(keepends=True))
        # TODO: we only extract those when outputting docbook, thats silly
        self.assertEqual({}, mkdb.SymbolAnnotations)

    # multiple annotations, multiline annotations, symbol-level ...


class OutputStruct(unittest.TestCase):

    def test_SimpleStructGetNormalized(self):
        res = mkdb.extract_struct_body('data', textwrap.dedent("""\
            struct data
            {
                int i;
                char *c;
            };
            """), False, True)
        self.assertEqual(textwrap.dedent("""\
            struct data {
                int i;
                char *c;
            };
            """), res)

    def test_SimpleTypedefStructGetNormalized(self):
        res = mkdb.extract_struct_body('data', textwrap.dedent("""\
            typedef struct _data
            {
                int i;
                char *c;
            } data;
            """), True, True)
        self.assertEqual(textwrap.dedent("""\
            typedef struct {
                int i;
                char *c;
            } data;
            """), res)

    def test_InternalStructNameIsNormalized(self):
        res = mkdb.extract_struct_body('data', textwrap.dedent("""\
            struct _data {
                int i;
                char *c;
            };
            """), False, True)
        self.assertEqual(textwrap.dedent("""\
            struct data {
                int i;
                char *c;
            };
            """), res)


if __name__ == '__main__':
    unittest.main()



# -*- python -*-
#
# gtk-doc - GTK DocBook documentation generator.
# Copyright (C) 2017  Stefan Sauer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#

from unittest import mock
import unittest

from gtkdoc import common


class TestUpdateFileIfChanged(unittest.TestCase):

    @mock.patch('os.path.exists')
    @mock.patch('os.rename')
    def test_NoOldFile(self, os_rename, os_path_exists):
        os_path_exists.return_value = False
        res = common.UpdateFileIfChanged('/old', '/new', False)
        os_rename.assert_called_with('/new', '/old')
        self.assertTrue(res)

    @mock.patch('os.path.exists')
    @mock.patch('builtins.open', mock.mock_open(read_data='bar'))
    @mock.patch('os.unlink')
    def test_FilesAreTheSame(self, os_unlink, os_path_exists):
        os_path_exists.return_value = True
        res = common.UpdateFileIfChanged('/old', '/new', False)
        os_unlink.assert_called_with('/new')
        self.assertFalse(res)


class TestGetModuleDocDir(unittest.TestCase):

    @mock.patch('subprocess.check_output')
    def test_ReturnsPath(self, subprocess_check_output):
        subprocess_check_output.return_value = '/usr'
        self.assertEqual(common.GetModuleDocDir('glib-2.0'), '/usr/share/gtk-doc/html')


class TestCreateValidSGMLID(unittest.TestCase):

    def test_AlreadyValid(self):
        self.assertEqual(common.CreateValidSGMLID('x'), 'x')

    def test_SpecialCharsBecomeDash(self):
        self.assertEqual(common.CreateValidSGMLID('x_ y'), 'x--y')

    def test_SpecialCharsGetRemoved(self):
        self.assertEqual(common.CreateValidSGMLID('x,;y'), 'xy')


if __name__ == '__main__':
    unittest.main()



# -*- python; coding: utf-8 -*-
#
# gtk-doc - GTK DocBook documentation generator.
# Copyright (C) 2009-2017  Stefan Sauer
#               2017  Jussi Pakkanen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#

import logging
import os
import sys
import subprocess

from . import config


def run_xsltproc(options, args):
    command = [config.xsltproc]
    # we could do "--path $PWD " to avoid needing rewriting entities that are
    # copied from the header into docs under xml
    if len(options.path):
        command += ['--path', ':'.join(options.path)]
    logging.info('running "%s"', ' '.join(command + args))
    pc = subprocess.Popen(command + args, stderr=subprocess.PIPE)
    (o, stde) = pc.communicate()
    with open('profile.txt', 'wb') as h:
        h.write(stde)
    return pc.returncode


def run(options):
    logging.info('options: %s', str(options.__dict__))

    module = options.args[0]
    document = options.args[1]

    if options.uninstalled:
        # this does not work from buiddir!=srcdir
        # we could try this
        # MAKE_SCRDIR=$(abs_srcdir) MAKE_BUILDDIR=$(abs_builddir) gtkdoc-mkpdf ...
        gtkdocdir = os.path.split(sys.argv[0])[0]
    else:
        gtkdocdir = os.path.join(config.datadir, 'gtk-doc/data')

    if config.dblatex != '':
        # extra options to consider
        # -I FIG_PATH
        # -V is useful for debugging
        # -T db2latex : different style
        # -d : keep transient files (for debugging)
        # -P abc.def=$quiet : once the stylesheets have a quiet mode
        # -x "--path /path/to/more/files"
        # xsltproc is already called with --xinclude
        # does not work: --xslt-opts "--path $searchpath --nonet $@"
        dblatex_options = ['-o', module + '.pdf']
        for i in options.imgdir:
            dblatex_options += ['-I', i]
        if len(options.path):
            dblatex_options += ['-x', '--path ' + ':'.join(options.path)]
        dblatex_options.append(document)
        if not options.verbose:
            pc = subprocess.Popen([config.dblatex, '--help'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            (stdo, stde) = pc.communicate()
            if b'--quiet' in stdo or b'--quiet' in stde:
                dblatex_options = ['--quiet'] + dblatex_options
        dbcmd = [config.dblatex] + dblatex_options
        pc = subprocess.Popen(dbcmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        (stde, _) = pc.communicate()
        for line in stde.decode('utf-8').split('\n'):
            if not line.strip():
                continue
            if 'programlisting or screen' in line:
                continue
            # This happens when dblatex has no support for some special chars
            if 'Missing character' in line:
                continue
            print(line)
        res = pc.returncode
    elif config.fop != '':
        if options.verbose:
            quiet = '0'
        else:
            quiet = '1'
        res = run_xsltproc(options, ['--nonet',
                                     '--xinclude',
                                     '--stringparam',
                                     'gtkdoc.bookname',
                                     module,
                                     '--stringparam',
                                     'gtkdoc.version',
                                     config.version,
                                     '--stringparam',
                                     'chunk.quietly',
                                     quiet,
                                     '--stringparam',
                                     'chunker.output.quiet',
                                     quiet,
                                     module,
                                     document,
                                     '-o',
                                     module + '.fo',
                                     gtkdocdir + '/gtk-doc-fo.xsl',
                                     document])
        # TODO: fop dies too easily :(
        # res = subprocess.call([config.fop, module + '.fo', module + '.pdf'))
        fname = module + '.fo'
        if os.path.exists(fname):
            os.unlink(fname)
    else:
        print("dblatex or fop must be installed to use gtkdoc-mkpdf.")
        res = 1

    with open('pdf.stamp', 'w') as h:
        h.write('timestamp')
    return res



# -*- python -*-
#
# gtk-doc - GTK DocBook documentation generator.
# Copyright (C) 2018  Stefan Sauer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#

import argparse
import textwrap
import unittest

from parameterized import parameterized

from gtkdoc import scan


BASIC_TYPES = [
    "char",
    "signed char",
    "unsigned char",
    "short",
    "signed short",
    "unsigned short",
    "long",
    "signed long",
    "unsigned long",
    "int",
    "signed int",
    "unsigned int",
    "short int",
    "signed short int",
    "unsigned short int",
    "long int",
    "signed long int",
    "unsigned long int",
    "signed",
    "unsigned",
    "float",
    "double",
    "long double",
    "enum e",
    "struct s",
    "union u",
]

BASIC_TYPES_WITH_VOID = ['void'] + BASIC_TYPES

INLINE_MODIFIERS = [('g_inline', 'G_INLINE_FUNC'), ('static_inline', 'static inline')]


class ScanHeaderContentTestCase(unittest.TestCase):
    """Baseclass for the header scanner tests."""

    def setUp(self):
        self.decls = []
        self.types = []
        self.options = argparse.Namespace(
            deprecated_guards='GTKDOC_TESTER_DISABLE_DEPRECATED',
            ignore_decorators='',
            rebuild_types=False)
        scan.InitScanner(self.options)

    def scanHeaderContent(self, content):
        return scan.ScanHeaderContent(content, self.decls, self.types,
                                      self.options)

    def assertNoDeclFound(self, slist):
        self.assertEqual([], slist)
        self.assertEqual([], self.decls)
        self.assertEqual([], self.types)

    def assertNothingFound(self, slist, doc_comments):
        self.assertEqual({}, doc_comments)
        self.assertNoDeclFound(slist)


class ScanHeaderContent(ScanHeaderContentTestCase):
    """Test generic scanner behaviour."""

    def test_EmptyInput(self):
        slist, doc_comments = self.scanHeaderContent([])
        self.assertNothingFound(slist, doc_comments)

    def test_IgnoresOneLineComments(self):
        slist, doc_comments = self.scanHeaderContent(['/* test */'])
        self.assertNothingFound(slist, doc_comments)

    def test_IgnoresOneLineCommentsDoubleStar(self):
        slist, doc_comments = self.scanHeaderContent(['/** test */'])
        self.assertNothingFound(slist, doc_comments)

    def test_FindsDocComment(self):
        slist, doc_comments = self.scanHeaderContent("""\
            /**
             * Symbol:
             */""".splitlines(keepends=True))
        self.assertEqual(1, len(doc_comments))
        self.assertIn('symbol', doc_comments)

    def test_DocDoesNotChangeSlistDeclAndTypes(self):
        slist, doc_comments = self.scanHeaderContent("""\
            /**
             * Symbol:
             */""".splitlines(keepends=True))
        self.assertNoDeclFound(slist)

    def test_SkipInternalHeaders(self):
        header = textwrap.dedent("""\
            /* < private_header > */
            /**
             * symbol:
             */
            void symbols(void);""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertNothingFound(slist, doc_comments)

    def test_SkipSymbolWithPreprocessor(self):
        slist, doc_comments = self.scanHeaderContent("""\
            #ifndef __GTK_DOC_IGNORE__
            extern int bug_512565(void);
            #endif""".splitlines(keepends=True))
        self.assertNoDeclFound(slist)

    def test_AddDeprecatedFlagForSymbolsWithinDeprecationGuards(self):
        header = textwrap.dedent("""\
            #ifndef GTKDOC_TESTER_DISABLE_DEPRECATED
            /**
             * SYMBOL:
             *
             * Deprecated: 1.1. Use NEW_SYMBOL instead.
             */
            #define SYMBOL "value"
            #endif /* GTKDOC_TESTER_DISABLE_DEPRECATED */""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertEqual(1, len(self.decls))
        self.assertIn('<DEPRECATED/>', self.decls[0])

    def test_NoDeprecatedFlagForSymbolsOutsideDeprecationGuards(self):
        header = textwrap.dedent("""\
            #ifndef GTKDOC_TESTER_DISABLE_DEPRECATED
            /**
             * SYMBOL1:
             *
             * Deprecated: 1.1. Use NEW_SYMBOL1 instead.
             */
            #define SYMBOL1 "value"
            #endif /* GTKDOC_TESTER_DISABLE_DEPRECATED */
            /**
             * SYMBOL2:
             */
            #define SYMBOL2 "value"
            """)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertEqual(2, len(self.decls))
        self.assertNotIn('<DEPRECATED/>', self.decls[1])


class ScanHeaderContentEnum(ScanHeaderContentTestCase):
    """Test parsing of enum declarations."""

    def assertDecl(self, name, decl, slist):
        self.assertEqual([name], slist)
        d = '<ENUM>\n<NAME>%s</NAME>\n%s</ENUM>\n' % (name, decl)
        self.assertEqual([d], self.decls)
        self.assertEqual([], self.types)

    def test_FindsEnum(self):
        header = textwrap.dedent("""\
            enum data {
              TEST,
            };""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('data', header, slist)

    def test_FindsTypedefEnum(self):
        header = textwrap.dedent("""\
            typedef enum {
              ENUM
            } Data;""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('Data', header, slist)

    def test_HandleEnumWithDeprecatedMember(self):
        header = textwrap.dedent("""\
            enum data {
              TEST_A,
            #ifndef GTKDOC_TESTER_DISABLE_DEPRECATED
              TEST_B,
            #endif
              TEST_C
            };""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('data', header, slist)

    def test_HandleDeprecatedInMemberName(self):
        header = textwrap.dedent("""\
            typedef enum {
              VAL_DEFAULT,
              VAL_DEPRECATED,
            } Data;""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('Data', header, slist)

    def test_HandleDeprecatedMemberDecorator(self):
        header = textwrap.dedent("""\
            typedef enum {
              VAL_DEFAULT,
              OTHER_VAL MY_DEPRECATED_FOR(VAL_DEFAULT),
            } Data;""")
        expected = textwrap.dedent("""\
            typedef enum {
              VAL_DEFAULT,
              OTHER_VAL,
            } Data;""")
        self.options.ignore_decorators = 'MY_DEPRECATED_FOR()'
        scan.InitScanner(self.options)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('Data', expected, slist)

    def test_HandleDeprecatedDecorator(self):
        header = textwrap.dedent("""\
            typedef enum {
              VAL_DEFAULT,
              OTHER_VAL,
            } Data MY_DEPRECATED_FOR(OtherEnum);""")
        expected = textwrap.dedent("""\
            <DEPRECATED/>
            typedef enum {
              VAL_DEFAULT,
              OTHER_VAL,
            } Data;""")
        self.options.ignore_decorators = 'MY_DEPRECATED_FOR()'
        scan.InitScanner(self.options)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('Data', expected, slist)


class ScanHeaderContentFunctions(ScanHeaderContentTestCase):
    """Test parsing of function declarations."""

    def assertDecl(self, name, ret, params, slist):
        self.assertEqual([name], slist)
        d = '<FUNCTION>\n<NAME>%s</NAME>\n<RETURNS>%s</RETURNS>\n%s\n</FUNCTION>\n' % (name, ret, params)
        self.assertEqual([d], self.decls)
        self.assertEqual([], self.types)

    def test_FindsFunctionVoid(self):
        header = 'void func();'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', 'void', '', slist)

    def test_IgnoresInternalFunction(self):
        slist, doc_comments = self.scanHeaderContent([
            'void _internal(void);'
        ])
        self.assertNoDeclFound(slist)

    @parameterized.expand(INLINE_MODIFIERS)
    def test_IgnoresInternalInlineVoidFunction(self, _, modifier):
        header = textwrap.dedent("""\
            %s void _internal(void) {
            }""" % modifier)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertNoDeclFound(slist)

    @parameterized.expand(INLINE_MODIFIERS)
    def test_IgnoresInternalInlineFunction(self, _, modifier):
        header = textwrap.dedent("""\
            %s int _internal(int a) {
              return a + a; }""" % modifier)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertNoDeclFound(slist)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES_WITH_VOID])
    def test_HandlesReturnValue(self, _, ret_type):
        header = '%s func(void);' % ret_type
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', ret_type, 'void', slist)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES_WITH_VOID])
    def test_HandlesReturnValuePtr(self, _, ret_type):
        header = '%s* func(void);' % ret_type
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', ret_type + ' *', 'void', slist)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES_WITH_VOID])
    def test_HandlesReturnValueConstPtr(self, _, ret_type):
        header = 'const %s* func(void);' % ret_type
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', 'const ' + ret_type + ' *', 'void', slist)

    # TODO: also do parametrized?
    def test_FindsFunctionConstCharPtConstPtr_Void(self):
        header = 'const char* const * func(void);'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', 'const char * const *', 'void', slist)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES])
    def test_HandlesParameter(self, _, param_type):
        header = 'void func(%s);' % param_type
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', 'void', param_type, slist)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES])
    def test_HandlesNamedParameter(self, _, param_type):
        header = 'void func(%s a);' % param_type
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', 'void', param_type + ' a', slist)

    def test_HandlesMultipleParameterd(self):
        header = 'int func(char c, long l);'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', 'int', 'char c, long l', slist)

    def test_FindsFunctionStruct_Void_WithLinebreakAfterRetType(self):
        header = textwrap.dedent("""\
            struct ret *
            func (void);""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'struct ret *', 'void', slist)

    def test_FindsFunctionStruct_Void_WithLinebreakAfterFuncName(self):
        header = textwrap.dedent("""\
            struct ret * func
            (void);""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'struct ret *', 'void', slist)

    def test_FindsFunctionVoid_Int_WithLinebreakAfterParamType(self):
        header = textwrap.dedent("""\
            void func (int
              a);""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'void', 'int a', slist)

    @parameterized.expand(INLINE_MODIFIERS)
    def test_FindsInlineFunction(self, _, modifier):
        header = textwrap.dedent("""\
            %s void func (void)
            {
            }
            """ % modifier)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'void', 'void', slist)

    @parameterized.expand(INLINE_MODIFIERS)
    def test_FindsInlineFunctionWithNewlineAfterType(self, _, modifier):
        header = textwrap.dedent("""\
            %s void
            func (void)
            {
            }
            """ % modifier)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'void', 'void', slist)

    @parameterized.expand(INLINE_MODIFIERS)
    def test_FindsInlineFunctionWithConditionalBody(self, _, modifier):
        header = textwrap.dedent("""\
            %s int
            func (int a)
            {
            #if defined(__GNUC__) && (__GNUC__ >= 4) && defined(__OPTIMIZE__)
              return a;
            #else
              return 0;
            #endif
            }
            """ % modifier)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'int', 'int a', slist)

    @parameterized.expand(INLINE_MODIFIERS)
    def test_FindsInlineFunctionWithParenthesisName(self, _, modifier):
        header = textwrap.dedent("""\
            %s void
            (func) (void)
            {
            }
            """ % modifier)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'void', 'void', slist)


class ScanHeaderContentMacros(ScanHeaderContentTestCase):
    """Test parsing of macro declarations."""

    def assertDecl(self, name, decl, slist):
        self.assertEqual([name], slist)
        d = '<MACRO>\n<NAME>%s</NAME>\n%s</MACRO>\n' % (name, decl)
        self.assertEqual([d], self.decls)
        self.assertEqual([], self.types)

    def test_FindsMacroNumber(self):
        header = '#define FOO 1'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('FOO', header, slist)

    def test_FindsMacroExpression(self):
        header = '#define FOO (1 << 1)'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('FOO', header, slist)

    def test_FindsMacroFunction(self):
        header = '#define FOO(x) (x << 1)'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('FOO', header, slist)

    def test_IgnoresInternalMacro(self):
        slist, doc_comments = self.scanHeaderContent([
            '#define _INTERNAL (a) (a)'
        ])
        self.assertNoDeclFound(slist)

    def test_FindsDocCommentForDeprecationGuard(self):
        header = textwrap.dedent("""\
            /**
             * GTKDOC_TESTER_DISABLE_DEPRECATED:
             *
             * Documentation for a deprecation guard.
             */
            #define GTKDOC_TESTER_DISABLE_DEPRECATED 1""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertEqual(1, len(doc_comments))
        self.assertIn('gtkdoc_tester_disable_deprecated', doc_comments)
        self.assertDecl('GTKDOC_TESTER_DISABLE_DEPRECATED',
                        '#define GTKDOC_TESTER_DISABLE_DEPRECATED 1', slist)


class ScanHeaderContentStructs(ScanHeaderContentTestCase):
    """Test parsing of struct declarations."""

    def assertDecl(self, name, decl, slist):
        self.assertEqual([name], slist)
        d = '<STRUCT>\n<NAME>%s</NAME>\n%s</STRUCT>\n' % (name, decl)
        self.assertEqual([d], self.decls)
        self.assertEqual([], self.types)

    def test_FindsStruct(self):
        header = textwrap.dedent("""\
            struct data {
              int test;
            };""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('data', header, slist)

    def test_FindsTypedefStruct(self):
        header = textwrap.dedent("""\
            typedef struct {
              int test;
            } Data;""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('Data', header, slist)

    def test_HandleStructWithDeprecatedMember(self):
        header = textwrap.dedent("""\
            struct data {
              int test_a;
            #ifndef GTKDOC_TESTER_DISABLE_DEPRECATED
              int deprecated;
            #endif
              int test_b;
            };""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('data', header, slist)

    def test_IgnoresInternalStruct(self):
        header = 'struct _internal *x;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertNoDeclFound(slist)

    def test_IgnoresPrivateStruct(self):
        header = 'struct _x;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertNoDeclFound(slist)

    def test_OpaqueStructTypedefGeneratesEmptyDecl(self):
        header = 'typedef struct _data data;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('data', '', slist)

    def test_OpaqueStructGeneratesEmptyDecl(self):
        header = 'struct data;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('data', header, slist)

    def test_GetTitleFromGObjectClassStruct(self):
        header = textwrap.dedent("""\
            struct _GtkdocObjectClass {
              GObjectClass parent;
            };""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertIn('<TITLE>GtkdocObject</TITLE>', slist)

    def test_DeprecatedDecorator(self):
        header = textwrap.dedent("""\
            typedef struct {
              int x;
            } Data MY_DEPRECATED_FOR(OtherStruct);""")
        expected = textwrap.dedent("""\
            <DEPRECATED/>
            typedef struct {
              int x;
            } Data;""")
        self.options.ignore_decorators = 'MY_DEPRECATED_FOR()'
        scan.InitScanner(self.options)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('Data', expected, slist)

    def test_DeprecatedOpaqueStructTypedef(self):
        header = 'typedef struct _data data MY_DEPRECATED_FOR(OtherData);'
        expected = '<DEPRECATED/>\n'
        self.options.ignore_decorators = 'MY_DEPRECATED_FOR()'
        scan.InitScanner(self.options)
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('data', expected, slist)


class ScanHeaderContentUnions(ScanHeaderContentTestCase):
    """Test parsing of union declarations."""

    def assertDecl(self, name, decl, slist):
        self.assertEqual([name], slist)
        d = '<UNION>\n<NAME>%s</NAME>\n%s</UNION>\n' % (name, decl)
        self.assertEqual([d], self.decls)
        self.assertEqual([], self.types)

    def test_FindsUnion(self):
        header = textwrap.dedent("""\
            union data {
              int i;
              float f;
            };""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('data', header, slist)

    def test_FindsTypedefUnion(self):
        header = textwrap.dedent("""\
            typedef union {
              int i;
              float f;
            } Data;""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('Data', header, slist)

    def test_IgnoresInternalUnion(self):
        header = 'union _internal *x;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertNoDeclFound(slist)

    def test_IgnoresPrivateUnion(self):
        header = 'union _x;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertNoDeclFound(slist)

    def test_OpaqueUnionTypedefGeneratesEmptyDecl(self):
        header = 'typedef union _data data;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('data', '', slist)

    def test_OpaqueUnionGeneratesEmptyDecl(self):
        header = 'union data;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('data', header, slist)


class ScanHeaderContentUserFunction(ScanHeaderContentTestCase):
    """Test parsing of function pointer declarations."""

    def assertDecl(self, name, ret, params, slist):
        self.assertEqual([name], slist)
        d = '<USER_FUNCTION>\n<NAME>%s</NAME>\n<RETURNS>%s</RETURNS>\n%s</USER_FUNCTION>\n' % (name, ret, params)
        self.assertEqual([d], self.decls)
        self.assertEqual([], self.types)

    def test_FindsFunctionVoid(self):
        header = 'typedef void (*func)();'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', 'void', '', slist)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES_WITH_VOID])
    def test_HandlesReturnValue(self, _, ret_type):
        header = 'typedef %s (*func)(void);' % ret_type
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', ret_type, 'void', slist)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES_WITH_VOID])
    def test_HandlesReturnValuePtr(self, _, ret_type):
        header = 'typedef %s* (*func)(void);' % ret_type
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', ret_type + ' *', 'void', slist)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES_WITH_VOID])
    def test_HandlesReturnValueConstPtr(self, _, ret_type):
        header = 'typedef const %s* (*func)(void);' % ret_type
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', 'const ' + ret_type + ' *', 'void', slist)

    def test_FindsFunctionVoid_Int_WithLinebreakAfterTypedef(self):
        header = textwrap.dedent("""\
            typedef
            void (*func) (int a);""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'void', 'int a', slist)

    def test_FindsFunctionStruct_Void_WithLinebreakAfterRetType(self):
        header = textwrap.dedent("""\
            typedef struct ret *
            (*func) (void);""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'struct ret *', 'void', slist)

    # TODO: not found
    # def test_FindsFunctionStruct_Void_WithLinebreakAfterFuncName(self):
    #     header = textwrap.dedent("""\
    #         typedef struct ret * (*func)
    #         (void);""")
    #     slist, doc_comments = self.scanHeaderContent(
    #         header.splitlines(keepends=True))
    #     self.assertDecl('func', 'struct ret *', 'void', slist)

    def test_FindsFunctionVoid_Int_WithLinebreakAfterParamType(self):
        header = textwrap.dedent("""\
            typedef void (*func) (int
              a);""")
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('func', 'void', 'int a', slist)

    @parameterized.expand([('void', 'void'), ('const_int', 'const int')])
    def test_FindsFunctionPointerVar(self, _, ret_type):
        header = '%s (*func)();' % ret_type
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('func', ret_type, '', slist)


class ScanHeaderContentVariabless(ScanHeaderContentTestCase):
    """Test parsing of variable declarations."""

    def assertDecl(self, name, decl, slist):
        self.assertEqual([name], slist)
        d = '<VARIABLE>\n<NAME>%s</NAME>\n%s</VARIABLE>\n' % (name, decl)
        self.assertEqual([d], self.decls)
        self.assertEqual([], self.types)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES])
    def test_FindsExternVar(self, _, var_type):
        header = 'extern %s var;' % var_type
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('var', header, slist)

    @parameterized.expand([(t.replace(' ', '_'), t) for t in BASIC_TYPES])
    def test_FindsExternPtrVar(self, _, var_type):
        header = 'extern %s* var;' % var_type
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('var', header, slist)

    def test_FindsConstInt(self):
        header = 'const int var = 42;'
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('var', header, slist)

    def test_FindConstCharPtr(self):
        header = 'const char* var = "foo";'
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        self.assertDecl('var', header, slist)


class ScanHeaderContentTypedefs(ScanHeaderContentTestCase):
    """Test parsing of typedef declarations."""

    def assertDecl(self, name, decl, slist):
        self.assertEqual([name], slist)
        d = '<TYPEDEF>\n<NAME>%s</NAME>\n%s</TYPEDEF>\n' % (name, decl)
        self.assertEqual([d], self.decls)
        self.assertEqual([], self.types)

    def test_FindsTypedefStructPointer(self):
        header = 'typedef struct data *dataptr;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('dataptr', header, slist)

    def test_FindsTypedefUnionPointer(self):
        header = 'typedef union data *dataptr;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('dataptr', header, slist)

    def test_FindsTypedef(self):
        header = 'typedef unsigned int uint;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertDecl('uint', header, slist)

    def test_SkipsEnumTypedefUnderscoreName(self):
        header = 'typedef enum _data data;'
        slist, doc_comments = self.scanHeaderContent([header])
        self.assertNoDeclFound(slist)


class SeparateSubSections(ScanHeaderContentTestCase):

    def test_NoSymbolsGiveEmptyResult(self):
        liststr = scan.SeparateSubSections([], {})
        self.assertEqual('\n', liststr)

    def test_CreatesStandardSectionFromIsObjectMacro(self):
        header = textwrap.dedent("""\
            #define GTKDOC_IS_OBJECT(obj) (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GTKDOC_TYPE_OBJECT))
            void gtkdoc_object_function(void);
            """)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        liststr = scan.SeparateSubSections(slist, doc_comments)
        self.assertEqual(
            ['gtkdoc_object_function', '<SUBSECTION Standard>', 'GTKDOC_IS_OBJECT'],
            liststr.splitlines())

    def test_CreatesStandardSectionFromIsObjectClassMacro(self):
        header = textwrap.dedent("""\
            #define GTKDOC_IS_OBJECT_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), GTKDOC_TYPE_OBJECT))
            void gtkdoc_object_function(void);
            """)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        liststr = scan.SeparateSubSections(slist, doc_comments)
        self.assertEqual(
            ['gtkdoc_object_function', '<SUBSECTION Standard>', 'GTKDOC_IS_OBJECT_CLASS'],
            liststr.splitlines())

    def test_CreatesStandardSectionFromGetTypeFunction(self):
        header = textwrap.dedent("""\
            GType gtkdoc_object_get_type(void) G_GNUC_CONST;
            void gtkdoc_object_function(void);
            """)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        liststr = scan.SeparateSubSections(slist, doc_comments)
        self.assertEqual(
            ['gtkdoc_object_function', '<SUBSECTION Standard>', 'gtkdoc_object_get_type'],
            liststr.splitlines())

    def test_CreatesStandardSectionAllMacros(self):
        header = textwrap.dedent("""\
            #define GTKDOC_TYPE_OBJECT            (gtkdoc_object_get_type())
            #define GTKDOC_OBJECT(obj)            (G_TYPE_CHECK_INSTANCE_CAST((obj), GTKDOC_TYPE_OBJECT, GtkdocObject))
            #define GTKDOC_OBJECT_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST((klass),  GTKDOC_TYPE_OBJECT, GtkdocObjectClass))
            #define GTKDOC_IS_OBJECT(obj)         (G_TYPE_CHECK_INSTANCE_TYPE((obj), GTKDOC_TYPE_OBJECT))
            #define GTKDOC_IS_OBJECT_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE((klass),  GTKDOC_TYPE_OBJECT))
            #define GTKDOC_OBJECT_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS((obj),  GTKDOC_TYPE_OBJECT, GtkdocObjectClass))
            """)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        liststr = scan.SeparateSubSections(slist, doc_comments)
        self.assertEqual(
            ['<SUBSECTION Standard>', 'GTKDOC_IS_OBJECT', 'GTKDOC_IS_OBJECT_CLASS',
             'GTKDOC_OBJECT', 'GTKDOC_OBJECT_CLASS', 'GTKDOC_OBJECT_GET_CLASS',
             'GTKDOC_TYPE_OBJECT'],
            liststr.splitlines())

    def test_MovesSymbolIfUndocumented(self):
        header = textwrap.dedent("""\
            struct _GtkdocObject {
              GObject parent;
            };
            GType gtkdoc_object_get_type(void) G_GNUC_CONST;
            void gtkdoc_object_function(void);
            """)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        liststr = scan.SeparateSubSections(slist, doc_comments)
        self.assertEqual(
            ['gtkdoc_object_function', '<SUBSECTION Standard>', 'GtkdocObject', 'gtkdoc_object_get_type'],
            liststr.splitlines())

    def test_DoesNotMoveSymbolIfDocumented(self):
        header = textwrap.dedent("""\
            /**
             * GtkdocObject:
             *
             * instance data of gtk-doc unit test class
             */
            struct _GtkdocObject {
              GObject parent;
            };
            GType gtkdoc_object_get_type(void) G_GNUC_CONST;
            void gtkdoc_object_function(void);
            """)
        slist, doc_comments = self.scanHeaderContent(
            header.splitlines(keepends=True))
        liststr = scan.SeparateSubSections(slist, doc_comments)
        self.assertEqual(
            ['GtkdocObject', 'gtkdoc_object_function', '<SUBSECTION Standard>', 'gtkdoc_object_get_type'],
            liststr.splitlines())


class RemoveBracedContent(ScanHeaderContentTestCase):

    def test_OneLineFunctionBodyIsRemoved(self):
        decl = textwrap.dedent("""\
            static inline int function(int a) { return a + a; }""")
        (skip, decl) = scan.remove_braced_content(decl)
        self.assertEqual("static inline int function(int a);", decl)
        self.assertEqual(skip, False)

    def test_SimpleFunctionBodyIsRemoved(self):
        decl = textwrap.dedent("""\
            static inline int function(int a) {
              return a + a;
            }""")
        (skip, decl) = scan.remove_braced_content(decl)
        self.assertEqual("static inline int function(int a);", decl)
        self.assertEqual(skip, False)

    def test_SimpleFunctionWithNewlineBodyIsRemoved(self):
        decl = textwrap.dedent("""\
            static inline int function(int a)
            {
              return a + a;
            }""")
        (skip, decl) = scan.remove_braced_content(decl)
        self.assertEqual("static inline int function(int a);", decl)
        self.assertEqual(skip, False)

    def test_NestedFunctionBodyIsRemoved(self):
        decl = textwrap.dedent("""\
            static inline int function(int a) {
              if (a > 0) {
                return a + a;
              } else {
                return a - a;
              }
            }""")
        (skip, decl) = scan.remove_braced_content(decl)
        self.assertEqual("static inline int function(int a);", decl)
        self.assertEqual(skip, False)

    def test_NestedFunctionWithNewlinesBodyIsRemoved(self):
        decl = textwrap.dedent("""\
            static inline int function(int a)
            {
              if (a > 0)
              {
                return a + a;
              }
              else
              {
                return a - a;
              }
            }""")
        (skip, decl) = scan.remove_braced_content(decl)
        self.assertEqual("static inline int function(int a);", decl)
        self.assertEqual(skip, False)

    def test_SimpleFunctionWithParenthesisBodyIsRemoved(self):
        decl = textwrap.dedent("""\
            static inline int
            (function) (int a)
            {
              return a + a;
            }""")
        (skip, decl) = scan.remove_braced_content(decl)
        self.assertEqual("static inline int\n(function) (int a);", decl)
        self.assertEqual(skip, False)

    def test_FunctionWithMultilineParamsBodyIsRemoved(self):
        decl = textwrap.dedent("""\
            static inline int
            function (int a,
                      int b)
            {
              return a + b;
            }""")
        (skip, decl) = scan.remove_braced_content(decl)
        self.assertEqual(
            "static inline int\nfunction (int a,\n          int b);", decl)
        self.assertEqual(skip, False)


if __name__ == '__main__':
    from gtkdoc import common
    common.setup_logging()

    unittest.main()

    # t = RemoveBracedContent()
    # t.setUp()
    # t.test_NestedFunctionBodyIsRemoved()



# -*- python; coding: utf-8 -*-
#
# gtk-doc - GTK DocBook documentation generator.
# Copyright (C) 2018  Stefan Sauer
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#

import logging
import textwrap
import unittest

from anytree import PreOrderIter
from lxml import etree
from parameterized import parameterized

from gtkdoc import mkhtml2


class TestChunking(unittest.TestCase):

    # def setUp(self):
    #     logging.basicConfig(
    #         level=logging.INFO,
    #         format='%(asctime)s:%(filename)s:%(funcName)s:%(lineno)d:%(levelname)s:%(message)s')

    def test_chunk_only_root_gives_single_chunk(self):
        root = etree.XML('<book />')
        files = mkhtml2.chunk(root, 'test')
        self.assertEqual('book', files.name)
        self.assertEqual(0, len(files.descendants))

    def test_chunk_single_chapter_gives_two_chunks(self):
        root = etree.XML('<book><chapter /></book>')
        files = mkhtml2.chunk(root, 'test')
        descendants = [f for f in files.descendants if f.anchor is None]
        logging.info('descendants : %s', str(descendants))
        self.assertEqual(1, len(descendants))

    def test_chunk_first_sect1_is_inlined(self):
        root = etree.XML('<book><chapter><sect1 /></chapter></book>')
        files = mkhtml2.chunk(root, 'test')
        descendants = [f for f in files.descendants if f.anchor is None]
        logging.info('descendants : %s', str(descendants))
        self.assertEqual(1, len(descendants))

    def test_chunk_second_sect1_is_not_inlined(self):
        root = etree.XML('<book><chapter><sect1 /><sect1 /></chapter></book>')
        files = mkhtml2.chunk(root, 'test')
        descendants = [f for f in files.descendants if f.anchor is None]
        logging.info('descendants : %s', str(descendants))
        self.assertEqual(2, len(descendants))


class TestDataExtraction(unittest.TestCase):

    # def setUp(self):
    #     logging.basicConfig(
    #         level=logging.INFO,
    #         format='%(asctime)s:%(filename)s:%(funcName)s:%(lineno)d:%(levelname)s:%(message)s')

    def chunk_db(self, xml):
        root = etree.XML(xml)
        files = mkhtml2.chunk(root, 'test')
        return [f for f in PreOrderIter(files) if f.anchor is None]

    def test_extract_ids(self):
        files = self.chunk_db('<book><chapter id="chap1"></chapter></book>')
        links = {}
        mkhtml2.add_id_links_and_titles(files, links)
        self.assertIn('chap1', links)

    def test_extract_titles(self):
        files = self.chunk_db('<book><chapter id="chap1"><title>Intro</title></chapter></book>')
        links = {}
        mkhtml2.add_id_links_and_titles(files, links)
        self.assertIn('chap1', mkhtml2.titles)
        self.assertEqual('Intro', mkhtml2.titles['chap1']['title'])
        self.assertEqual('chapter', mkhtml2.titles['chap1']['tag'])

    def test_extract_glossaries(self):
        files = self.chunk_db(textwrap.dedent("""\
            <book>
              <glossary id="glossary">
                <glossentry>
                  <glossterm><anchor id="glossterm-API"/>API</glossterm>
                  <glossdef>
                    <para>Application Programming Interface</para>
                  </glossdef>
                </glossentry>
              </glossary>
            </book>"""))
        mkhtml2.build_glossary(files)
        self.assertIn('API', mkhtml2.glossary)
        self.assertEqual('Application Programming Interface', mkhtml2.glossary['API'])


class TestDevhelp(unittest.TestCase):

    xml_minimal = textwrap.dedent("""\
      <book>
        <chapter id="chap1"><title>Intro</title></chapter>
      </book>""")

    xml_basic = textwrap.dedent("""\
      <book>
        <bookinfo>
          <title>test Reference Manual</title>
          <releaseinfo>
            The latest version of this documentation can be found on-line at
            <ulink role="online-location" url="http://www.example.com/tester/index.html">online-site</ulink>.
          </releaseinfo>
        </bookinfo>
        <chapter id="chap1"><title>Intro</title></chapter>
      </book>""")

    # TODO: need one with multiple conditions
    xml_full = textwrap.dedent("""\
      <book>
        <bookinfo>
          <title>test Reference Manual</title>
        </bookinfo>
        <chapter id="chap1">
          <title>Reference</title>
          <refentry id="GtkdocObject">
            <refmeta>
              <refentrytitle role="top_of_page" id="GtkdocObject.top_of_page">GtkdocObject</refentrytitle>
              <refmiscinfo>TESTER Library</refmiscinfo>
            </refmeta>
            <refnamediv>
              <refname>GtkdocObject</refname>
              <refpurpose>class for gtk-doc unit test</refpurpose>
            </refnamediv>
            <refsect1 id="GtkdocObject.functions" role="functions_proto">
              <title role="functions_proto.title">Functions</title>
            </refsect1>
            <refsect1 id="GtkdocObject.functions_details" role="details">
              <title role="details.title">Functions</title>
              <refsect2 id="gtkdoc-object-new" role="function" condition="since:0.1">
                <title>gtkdoc_object_new&#160;()</title>
                <refsect3 id="GtkdocObjectClass.members" role="struct_members">
                  <title>Members</title>
                  <informaltable role="struct_members_table" pgwide="1" frame="none">
                    <tgroup cols="3">
                      <colspec colname="struct_members_name" colwidth="300px"/>
                      <colspec colname="struct_members_description"/>
                      <colspec colname="struct_members_annotations" colwidth="200px"/>
                      <tbody>
                        <row role="member">
                          <entry role="struct_member_name">
                            <para>
                              GtkdocObjectClass
                              <structfield id="GtkdocObjectClass.parent">parent</structfield>;
                            </para>
                          </entry>
                          <entry role="struct_member_description">
                            <para>parent/</para>
                          </entry>
                          <entry role="struct_member_annotations"></entry>
                        </row>
                      </tbody>
                    </tgroup>
                  </informaltable>
                </refsect3>
              </refsect2>
            </refsect1>
            <refsect1 id="tester-GtkdocTypes.other_details" role="details">
              <title role="details.title">Types and Values</title>
              <refsect2 id="GtkdocEnum" role="enum">
                <title>enum GtkdocEnum</title>
                <indexterm zone="GtkdocEnum">
                  <primary>GtkdocEnum</primary>
                </indexterm>
                <para>Enum values for the <link linkend="GtkdocEnum"><type>GtkdocEnum</type>
                  </link> type.</para>
                <refsect3 id="GtkdocEnum.members" role="enum_members">
                  <title>Members</title>
                  <informaltable role="enum_members_table" pgwide="1" frame="none">
                    <tgroup cols="3">
                      <colspec colname="enum_members_name" colwidth="300px"/>
                      <colspec colname="enum_members_description"/>
                      <colspec colname="enum_members_annotations" colwidth="200px"/>
                      <tbody>
                        <row role="constant">
                          <entry role="enum_member_name">
                            <para id="GTKDOC-ENUM-V1:CAPS">GTKDOC_ENUM_V1</para>
                          </entry>
                          <entry role="enum_member_description">
                            <para>first</para>
                          </entry>
                          <entry role="enum_member_annotations"></entry>
                        </row>
                        <row role="constant">
                          <entry role="enum_member_name">
                            <para id="GTKDOC-ENUM-V2:CAPS">GTKDOC_ENUM_V2</para>
                          </entry>
                          <entry role="enum_member_description">
                            <para>second Since: 0.5</para>
                          </entry>
                          <entry role="enum_member_annotations"></entry>
                        </row>
                      </tbody>
                    </tgroup>
                  </informaltable>
                </refsect3>
              </refsect2>
            </refsect1>
          </refentry>
        </chapter>
      </book>""")

    # def setUp(self):
    #     logging.basicConfig(
    #         level=logging.INFO,
    #         format='%(asctime)s:%(filename)s:%(funcName)s:%(lineno)d:%(levelname)s:%(message)s')

    def convert(self, xml):
        root = etree.XML(xml)
        files = mkhtml2.chunk(root, 'test')
        files = [f for f in PreOrderIter(files) if f.anchor is None]
        mkhtml2.add_id_links_and_titles(files, {})
        return '\n'.join(mkhtml2.create_devhelp2_content('test', root, files))

    def test_create_devhelp_has_minimal_structure(self):
        devhelp = self.convert(self.xml_minimal)
        self.assertIn('<book xmlns', devhelp)
        self.assertIn('<chapters', devhelp)
        self.assertIn('<functions', devhelp)

    def test_create_devhelp_without_bookinfo(self):
        devhelp = self.convert(self.xml_minimal)
        self.assertNotIn('online', devhelp)

    def test_create_devhelp_with_bookinfo(self):
        devhelp = self.convert(self.xml_basic)
        self.assertIn('online="http://www.example.com/tester/index.html"', devhelp)
        self.assertIn('title="test Reference Manual"', devhelp)

    def test_create_devhelp_with_refentry_has_chapters(self):
        devhelp = self.convert(self.xml_full)
        self.assertIn('<sub name="Reference" link="chap1.html">', devhelp)
        self.assertIn('<sub name="GtkdocObject" link="GtkdocObject.html"/>', devhelp)

    def test_create_devhelp_with_refentry_has_keywords(self):
        devhelp = self.convert(self.xml_full)
        self.assertIn(
            '<keyword type="function" name="gtkdoc_object_new ()" '
            'link="GtkdocObject.html#gtkdoc-object-new" since="0.1"/>',
            devhelp)

    def test_create_devhelp_with_refesect3_has_member_keywords(self):
        devhelp = self.convert(self.xml_full)
        self.assertIn(
            '<keyword type="member" name="GtkdocObjectClass.parent" '
            'link="GtkdocObject.html#GtkdocObjectClass.parent"/>',
            devhelp)

    def test_create_devhelp_with_refesect3_has_constant_keywords(self):
        devhelp = self.convert(self.xml_full)
        self.assertIn(
            '<keyword type="constant" name="GTKDOC_ENUM_V1" '
            'link="GtkdocObject.html#GTKDOC-ENUM-V1:CAPS"/>',
            devhelp)


class TestNavNodes(unittest.TestCase):

    # def setUp(self):
    #     logging.basicConfig(
    #         level=logging.INFO,
    #         format='%(asctime)s:%(filename)s:%(funcName)s:%(lineno)d:%(levelname)s:%(message)s')

    def chunk_db(self, xml):
        root = etree.XML(xml)
        files = mkhtml2.chunk(root, 'test')
        return [f for f in PreOrderIter(files) if f.anchor is None]

    def test_nav_nodes_contains_home(self):
        files = self.chunk_db(textwrap.dedent("""\
            <book>
            </book>"""))
        nav = mkhtml2.generate_nav_nodes(files, files[0])
        self.assertEqual(1, len(nav))
        self.assertIn('nav_home', nav)

    def test_nav_nodes_contains_up_and_prev(self):
        files = self.chunk_db(textwrap.dedent("""\
            <book>
              <chapter id="chap1"><title>Intro</title></chapter>
            </book>"""))
        nav = mkhtml2.generate_nav_nodes(files, files[1])
        self.assertEqual(3, len(nav))
        self.assertIn('nav_up', nav)
        self.assertIn('nav_prev', nav)
        self.assertNotIn('nav_next', nav)

    def test_nav_nodes_contains_next(self):
        files = self.chunk_db(textwrap.dedent("""\
            <book>
              <chapter id="chap1"><title>Intro</title></chapter>
              <chapter id="chap2"><title>Content</title></chapter>
            </book>"""))
        nav = mkhtml2.generate_nav_nodes(files, files[1])
        self.assertEqual(4, len(nav))
        self.assertIn('nav_next', nav)


class TestConverter(unittest.TestCase):

    xml_book_beg = textwrap.dedent("""\
        <book>
          <bookinfo>
            <title>test Reference Manual</title>
          </bookinfo>""")
    xml_book_end = textwrap.dedent("""\
        </book>""")
    xml_book = '\n'.join([xml_book_beg, xml_book_end])

    xml_book_preface = '\n'.join([
        xml_book_beg,
        '  <preface id="intro"><title>Intro</title></preface>',
        xml_book_end])

    xml_book_chapter = '\n'.join([
        xml_book_beg,
        '  <chapter id="chap1"><title>Intro</title></chapter>',
        xml_book_end])

    xml_book_reference = '\n'.join([
        xml_book_beg,
        '  <reference id="part.i"><title>Reference</title></reference>',
        xml_book_end])

    xml_book_part_chapter = '\n'.join([
        xml_book_beg,
        textwrap.dedent("""\
          <part label="1" id="part.i">
            <title>Overview</title>
            <chapter id="chap1"><title>Intro</title></chapter>
          </part>"""),
        xml_book_end])

    # 2 sections since the first one is not chunked
    xml_book_chapter_secton = '\n'.join([
        xml_book_beg,
        textwrap.dedent("""\
          <chapter id="chap1">
            <title>Overview</title>
            <section id="s1"><title>Intro</title></section>
            <section id="s2"><title>getting started</title></section>
          </chapter>"""),
        xml_book_end])
    xml_book_chapter_sect1 = xml_book_chapter_secton.replace('section', 'sect1')

    xml_book_chapter_refentry_beg = '\n'.join([
        xml_book_beg,
        textwrap.dedent("""\
          <chapter id="chap1">
            <title>Reference</title>
            <refentry id="GtkdocObject">
              <refmeta>
                <refentrytitle role="top_of_page" id="GtkdocObject.top_of_page">GtkdocObject</refentrytitle>
              </refmeta>""")])

    xml_book_chapter_refentry_end = '\n'.join([
        textwrap.dedent("""\
            </refentry>
          </chapter>"""),
        xml_book_end])
    xml_book_chapter_refentry = '\n'.join([xml_book_chapter_refentry_beg, xml_book_chapter_refentry_end])

    xml_book_index_empty = '\n'.join([
        xml_book_beg,
        '  <index id="api-index-full"><title>API Index</title></index>',
        xml_book_end])

    xml_book_index = '\n'.join([
        xml_book_beg,
        textwrap.dedent("""\
          <index id="api-index-full">
            <title>API Index</title>
            <indexdiv id="api-index-full">
              <indexdiv><title>O</title>
                <!-- we get a link warning, since we we don't include the 'refentry' -->
                <!--indexentry>
                  <primaryie linkends="gtkdoc-object-new">
                    <link linkend="gtkdoc-object-new">gtkdoc_object_new</link>,
                    function in <link linkend="GtkdocObject">GtkdocObject</link>
                  </primaryie>
                </indexentry-->
              </indexdiv>
            </indexdiv>
          </index>"""),
        xml_book_end])

    xml_book_glossary_empty = '\n'.join([
        xml_book_beg,
        '  <glossary id="glossary"><title>Glossary</title></glossary>',
        xml_book_end])

    xml_book_glossary = '\n'.join([
        xml_book_beg,
        textwrap.dedent("""\
          <glossary id="glossary">
            <title>Glossary</title>
            <glossdiv><title>A</title>
              <glossentry>
                <glossterm><anchor id="glossterm-API"/>API</glossterm>
                <glossdef>
                  <para>Application Programming Interface</para>
                </glossdef>
              </glossentry>
            </glossdiv>
          </glossary>"""),
        xml_book_end])

    # def setUp(self):
    #     logging.basicConfig(
    #         level=logging.INFO,
    #         format='%(asctime)s:%(filename)s:%(funcName)s:%(lineno)d:%(levelname)s:%(message)s')

    def convert(self, xml, ix):
        root = etree.XML(xml)
        files = mkhtml2.chunk(root, 'test')
        nodes = [f for f in PreOrderIter(files) if f.anchor is None]
        return '\n'.join(mkhtml2.convert_content('test', nodes, nodes[ix], 'c'))

    @parameterized.expand([
        ('book', (xml_book, 0)),
        ('preface', (xml_book_preface, 1)),
        ('reference', (xml_book_reference, 1)),
        ('chapter', (xml_book_chapter, 1)),
        ('part', (xml_book_part_chapter, 1)),
        ('section', (xml_book_chapter_secton, 2)),
        ('sect1', (xml_book_chapter_sect1, 2)),
        ('refentry', (xml_book_chapter_refentry, 2)),
        ('index', (xml_book_index, 1)),
        ('index_empty', (xml_book_index_empty, 1)),
        ('glossary', (xml_book_glossary, 1)),
        ('glossary_empty', (xml_book_glossary_empty, 1)),
    ])
    def test_convert_produces_html(self, _, params):
        html = self.convert(params[0], params[1])
        self.assertIn('<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">', html)
        self.assertIn('<html>', html)
        self.assertIn('</html>', html)

    def test_convert_book_has_title(self):
        html = self.convert(self.xml_book, 0)
        self.assertIn('<title>test Reference Manual</title>', html)

    def test_refnav_includes_normal_refsect1(self):
        xml = '\n'.join([
            self.xml_book_chapter_refentry_beg,
            textwrap.dedent("""\
                  <refsect1 id="GtkdocObject.description" role="desc">
                     <title role="desc.title">Description</title>
                  </refsect1>
                """),
            self.xml_book_chapter_refentry_end])
        html = self.convert(xml, 2)
        self.assertIn('class="shortcut">Description</a>', html)

    def test_refnav_skips_protos_refsect1(self):
        xml = '\n'.join([
            self.xml_book_chapter_refentry_beg,
            textwrap.dedent("""\
                  <refsect1 id="GtkdocObject.functions" role="functions_proto">
                    <title role="functions_proto.title">Functions</title>
                  </refsect1>
                """),
            self.xml_book_chapter_refentry_end])
        html = self.convert(xml, 2)
        self.assertNotIn('class="shortcut">Functions</a>', html)


if __name__ == '__main__':
    unittest.main()



version = "1.31"

# tools
dblatex = ''
fop = ''
pkg_config = '/usr/bin/pkg-config'
xsltproc = '/usr/bin/xsltproc'

# configured directories
prefix = '/usr/local'
datarootdir = "${prefix}/share".replace('${prefix}', prefix)
datadir = "${datarootdir}".replace('${datarootdir}', datarootdir)

exeext = ''
